/* $Copyright$
 *  Copyright (C) Fujitsu Siemens Computers GmbH 2001, 2002
 *  All rights reserved
 */
#ident "$Header$"
#ifndef WITHOUT_INCLUDES
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE_SUB
#  define MODULE
#endif
#ifndef PRECOMPILE
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/system.h>
#endif
#ifndef WITHOUT_INCLUDES
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#if LINUX_VERSION_CODE >= 0x020400
static DECLARE_WAIT_QUEUE_HEAD(WaitQ);
#else
#define vm_pgoff                        vm_offset
#define pci_resource_start(dev, i)      (dev->base_address[i])
static struct wait_queue *WaitQ = NULL;
#endif
#endif
#ident "$Header: $"
#ident "$Header$"
#ident "$Header: $"
EXPORT_NO_SYMBOLS;
static const int SUCCESS = 0;
static const char *COP_DEV = "cop";
static unsigned long copRomStart;
static unsigned char *pvirtRomStart;
static unsigned char *pSp2HostCommunication;
static unsigned char *pIntForAgent;
static const unsigned long COP_ROM_LEN = 0x1000;
static const unsigned int SNI_VENDOR_ID = 0x110A;
static const unsigned int COP_DEVICE_ID = 0x001D;
static const u8 COP_DEVICE_CONTROL_REGISTER = 0x04;
static const u8 COP_BASE_ADDRES_REGISTER0 = 0x10;
static const u8 COP_BASE_ADDRES_REGISTER1 = 0x14;
static const u8 COP_BASE_ADDRES_REGISTER2 = 0x18;
static const u8 COP_INTERRUPT_LINE_REGISTER = 0x3C;
static const u8 COP_INTERRUPT_PIN_REGISTER = 0x3D;
static short copDebug = 0;
MODULE_PARM(copDebug, "h");
static int deviceOpen = 0;
static int cop_major = 0;
static struct pci_dev *copDev = NULL;
static int cop_interrupt_received=0;
static int cop_open(struct inode *inode, struct file *file)
{
    if (copDebug) printk(KERN_DEBUG "cop :" "cop_open: (%p,%p)\n",inode, file);
    if(deviceOpen) {
        if (copDebug) printk(KERN_DEBUG "cop :" "cop_open: device is busy\n");
        return -EBUSY;
    }
    deviceOpen++;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static int cop_release(struct inode *inode, struct file *file)
{
    deviceOpen--;
    if (copDebug) printk(KERN_DEBUG "cop :" "cop_release: (%p,%p)\n",inode, file);
    MOD_DEC_USE_COUNT;
    return SUCCESS;
}
static void cop_vma_open(struct vm_area_struct *area) {
    if (copDebug) printk(KERN_DEBUG "cop :" "cop_vma_open\n");
    MOD_INC_USE_COUNT;
}
static void cop_vma_close(struct vm_area_struct *area) {
    if (copDebug) printk(KERN_DEBUG "cop :" "cop_vma_close\n");
    MOD_DEC_USE_COUNT;
}
static struct vm_operations_struct cop_vm_ops = {
    cop_vma_open,
    cop_vma_close,
};
static int cop_mmap(struct file *file, struct vm_area_struct *vma)
{
    int err;
    unsigned long virt_add = vma->vm_start;
    unsigned long vsize = vma->vm_end-vma->vm_start;
    if (copDebug) printk(KERN_DEBUG "cop :" "cop_mmap virt=0x%lx phys=0x%lx(0x%lx) size=0x%lx(0x%lx)\n",virt_add, copRomStart, (long unsigned int)vma->vm_pgoff, COP_ROM_LEN, vsize);
    if (vsize > COP_ROM_LEN) {
        if (copDebug) printk(KERN_DEBUG "cop :" "cop_mmap: vsize > COP_ROM_LEN\n");
        return -EINVAL;
    }
#if defined(REDHAT) && defined(ADVANCED) 
    err = remap_page_range( 
	    vma, 
            virt_add, 
            copRomStart, 
            vsize, 
            vma->vm_page_prot); 
#else
    err = remap_page_range(
            virt_add,
            copRomStart,
            vsize,
            vma->vm_page_prot);
#endif
    if (err) {
        if (copDebug) printk(KERN_DEBUG "cop :" "cop_mmap: remap_page_range failed: %d\n",err);
        return -EAGAIN;
    }
    if (vma->vm_ops) {
        if (copDebug) printk(KERN_DEBUG "cop :" "cop_mmap: vma->vm_ops != NULL\n");
        return -EINVAL;
    }
    vma->vm_ops = &cop_vm_ops;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static void cop_irq_handler(int irq,void *dev_id,struct pt_regs *regs)
{
        if (*pIntForAgent == 0x77)
        {
                *pIntForAgent=0x00;
        }
        *pSp2HostCommunication =(1<<0);
    wake_up(&WaitQ);
        cop_interrupt_received=1;
}
static int cop_ioctl(
                                struct inode *inode,
                                struct file *file,
                                unsigned int cmd,
                                unsigned long arg )
{
        long flags;
        if (copDebug) printk(KERN_DEBUG "cop :" "-->cop_ioctl(0x%x)\n",cmd);
        if (_IOC_TYPE(cmd) != 'c') {
                if (copDebug) printk(KERN_DEBUG "cop :" "  type != magic\n");
                return -EINVAL;
        }
        if (_IOC_NR(cmd) > 1) {
                if (copDebug) printk(KERN_DEBUG "cop :" "  nr > maxnr\n");
                return -EINVAL;
        }
        switch(cmd)
        {
                case _IO('c', 0):
                        save_flags(flags);
                        cli();
                        if (cop_interrupt_received )
                        {
                        }
                        else
                        {
                                restore_flags(flags);
                                interruptible_sleep_on(&WaitQ);
                                save_flags(flags);
                                cli();
                        }
                        cop_interrupt_received=0;
                        restore_flags(flags);
                        break;
                default:
                        if (copDebug) printk(KERN_DEBUG "cop :" " ioctl-cmd=%x not valid\n",cmd);
                        return -EINVAL;
        }
        return SUCCESS;
}
static struct file_operations cop_fops =
{
    .ioctl = cop_ioctl,
    .mmap = cop_mmap,
    .open = cop_open,
    .release = cop_release
};
int init_module(void)
{
        int result;
    if (! pci_present()) {
        if (copDebug) printk(KERN_DEBUG "cop :" "init_module: pcibios not present\n");
        return -ENODEV;
    }
    copDev = pci_find_device(SNI_VENDOR_ID, COP_DEVICE_ID,copDev);
    if (copDev == NULL)
        {
        if (copDebug) printk(KERN_DEBUG "cop :" "init_module: copernicus not present!\n");
                return -ENODEV;
    }
        copRomStart = pci_resource_start(copDev, 0) & 0xfffffff0;
        pvirtRomStart=ioremap(copRomStart,COP_ROM_LEN);
        if (pvirtRomStart)
        {
                pSp2HostCommunication =pvirtRomStart + 0x404;
                pIntForAgent = pvirtRomStart + 0x1C;
        }
        if (copDebug) printk(KERN_DEBUG "cop :" "Copernicus found connected to IRQ %d\n",copDev->irq);
        result=request_irq(copDev->irq,cop_irq_handler,SA_SHIRQ | SA_INTERRUPT,"cop",copDev);
        if (result < 0)
        {
                if (copDebug) printk(KERN_DEBUG "cop :" "Bad irq number or handler\n");
                return result;
        }
    if((cop_major=register_chrdev(cop_major, COP_DEV, &cop_fops)) < 0) {
        printk(KERN_ERR "cop: unable to get major %d\n", cop_major);
        return -EIO;
    }
    return SUCCESS;
}
void cleanup_module(void)
{
    int err;
    wake_up(&WaitQ);
        free_irq(copDev->irq,copDev);
    err = unregister_chrdev(cop_major, COP_DEV);
    if (err < 0) {
        printk(KERN_ERR "cop: unregister_chrdev(%i) failed!\n", cop_major);
    }
    copDev = NULL;
}